"""Functions for automatic regularization and weighting of training data."""

# adapted from MorrowChem's RSS routines.
from __future__ import annotations

import ast
import traceback
from contextlib import suppress
from typing import TYPE_CHECKING

import numpy as np
from scipy.spatial import ConvexHull, Delaunay

if TYPE_CHECKING:
    from ase import Atoms


def set_custom_sigma(
    atoms: list[Atoms],
    reg_minmax: list[tuple],
    isolated_atom_energies: dict | None = None,
    scheme: str = "linear-hull",
    energy_name: str = "REF_energy",
    force_name: str = "REF_forces",
    virial_name: str = "REF_virial",
    element_order: list | None = None,
    max_energy: float = 20.0,
    config_type_override: dict | None = None,
    retain_existing_sigma: bool = False,
) -> list[Atoms]:
    """
    Handle automatic regularisation based on distance to convex hull, amongst other things.

    TODO: Need to make sure this works for full multi-stoichiometry systems.

    Parameters
    ----------
    atoms: (list of ase.Atoms)
        List of atoms objects to set reg. for. Usually fitting database
    reg_minmax: (list of tuples)
        List of tuples of (min, max) values for energy, force, virial sigmas
    scheme: (str)
        Method to use for regularization. Options are: linear_hull, volume-stoichiometry
        linear_hull: for single-composition system, use 2D convex hull (E, V)
        volume-stoichiometry: for multi-composition system, use 3D convex hull of (E, V, mole-fraction)
    energy_name: (str)
        Name of energy key in atoms.info
    force_name: (str)
        Name of force key in atoms.arrays
    virial_name: (str)
        Name of virial key in atoms.info
    isolated_atom_energies: (dict)
        Dictionary of isolated energies for each atomic number.
        Only needed for volume-x scheme e.g. {14: '-163.0', 8:'-75.0'}
        for SiO2
    element_order: (list)
        List of atomic numbers in order of choice (e.g. [42, 16] for MoS2)
    max_energy: (float)
        Ignore any structures with energy above hull greater than this value (eV)
    config_type_override: (dict)
        Give custom regularization for specific configuration types
    retain_existing_sigma: bool
        Whether to keep the current sigma values for specific configuration types.
        If set to True, existing sigma values for specific configurations will remain unchanged.

    e.g. reg_minmax = [(0.1, 1), (0.001, 0.1), (0.0316, 0.316), (0.0632, 0.632)]
    [(emin, emax), (semin, semax), (sfmin, sfmax), (ssvmin, svmax)]

    """
    sigs = [[], [], []]  # type: ignore
    atoms_modi = []

    if config_type_override is None:
        config_type_override = {
            "IsolatedAtom": (1e-4, 0.0, 0.0),
            "dimer": (0.1, 0.5, 0.0),
        }

    for at in atoms:
        for config_type, sigs_override in config_type_override.items():
            if at.info["config_type"] == config_type:
                at.info["energy_sigma"] = sigs_override[0]
                at.info["force_sigma"] = sigs_override[1]
                at.info["virial_sigma"] = sigs_override[2]
                atoms_modi.append(at)

        if at.info["config_type"] == "IsolatedAtom":
            at.calc = None  # TODO side-effect alert
            del at.info["force_sigma"]
            del at.info["virial_sigma"]
            continue

        if at.info["config_type"] == "dimer":
            with suppress(Exception):
                del at.info[virial_name]

            continue

    isolated_atom_energies = isolated_atom_energies or {}

    if scheme == "linear-hull":
        print("Regularising with linear hull")
        hull, points = get_convex_hull(atoms, energy_name=energy_name)
        get_e_distance_func = get_e_distance_to_hull

    elif scheme == "volume-stoichiometry":
        print("Regularising with 3D volume-mole fraction hull")
        if len(isolated_atom_energies) == 0:
            raise ValueError("Need to supply dictionary of isolated energies.")

        isolated_atom_energies = {
            ast.literal_eval(k) if isinstance(k, str) else k: v
            for k, v in isolated_atom_energies.items()
        }
        points = label_stoichiometry_volume(
            atoms, isolated_atom_energies, energy_name, element_order=element_order
        )  # label atoms with volume and mole fraction
        hull = calculate_hull_3d(points)  # calculate 3D convex hull
        get_e_distance_func = get_e_distance_to_hull_3d  # type: ignore

    points = {}
    for group in sorted(
        {  # check if set comprehension is running as supposed to
            at.info.get("rss_group")
            for at in atoms
            if not at.info.get("rss_nonperiodic")
        }
    ):
        points[group] = []

    for at in atoms:
        try:
            # skip non-periodic configs, volume is meaningless
            if at.info.get("rss_nonperiodic"):
                continue
            points[at.info.get("rss_group")].append(at)
        except Exception:
            pass

    for group, atoms_group in points.items():
        print("group:", group)

        for val in atoms_group:

            if retain_existing_sigma and "energy_sigma" in val.info:
                atoms_modi.append(val)
                continue

            de = get_e_distance_func(
                hull,
                val,
                energy_name=energy_name,
                isolated_atom_energies=isolated_atom_energies,
            )

            if de > max_energy:
                # don't even fit if too high
                continue
            if val.info["config_type"] not in config_type_override:
                if group == "initial":
                    sigs[0].append(reg_minmax[1][1])
                    sigs[1].append(reg_minmax[2][1])
                    sigs[2].append(reg_minmax[3][1])
                    val.info["energy_sigma"] = reg_minmax[1][1]
                    val.info["force_sigma"] = reg_minmax[2][1]
                    val.info["virial_sigma"] = reg_minmax[3][1]
                    atoms_modi.append(val)
                    continue

                if de <= reg_minmax[0][0]:
                    sigs[0].append(reg_minmax[1][0])
                    sigs[1].append(reg_minmax[2][0])
                    sigs[2].append(reg_minmax[3][0])
                    val.info["energy_sigma"] = reg_minmax[1][0]
                    val.info["force_sigma"] = reg_minmax[2][0]
                    val.info["virial_sigma"] = reg_minmax[3][0]
                    atoms_modi.append(val)

                elif de >= reg_minmax[0][1]:
                    sigs[0].append(reg_minmax[1][1])
                    sigs[1].append(reg_minmax[2][1])
                    sigs[2].append(reg_minmax[3][1])
                    val.info["energy_sigma"] = reg_minmax[1][1]
                    val.info["force_sigma"] = reg_minmax[2][1]
                    val.info["virial_sigma"] = reg_minmax[3][1]
                    atoms_modi.append(val)

                else:
                    [e, f, v] = piecewise_linear(
                        de,
                        [
                            (
                                0.1,
                                [reg_minmax[1][0], reg_minmax[2][0], reg_minmax[3][0]],
                            ),
                            (
                                1.0,
                                [reg_minmax[1][1], reg_minmax[2][1], reg_minmax[3][1]],
                            ),
                        ],
                    )
                    sigs[0].append(e)
                    sigs[1].append(f)
                    sigs[2].append(v)
                    val.info["energy_sigma"] = e
                    val.info["force_sigma"] = f
                    val.info["virial_sigma"] = v
                    atoms_modi.append(val)

    labels = ["E", "F", "V"]
    data_type = [np.array(sig) for sig in sigs]  # [e, f, v]

    for label, data in zip(labels, data_type):
        if len(data) == 0:
            print("No automatic regularisation performed (no structures requested)")
            continue
        if label == "E":
            # Report of the regularisation statistics
            print(f"Automatic regularisation statistics for {len(data)} structures:\n")
            print(
                "{:>20s}{:>20s}{:>20s}{:>20s}{:>20s}".format(
                    "", "Mean", "Std", "Nmin", "Nmax"
                )
            )
        print(
            f"{label:>20s}"
            f"{data.mean():>20.4f}"
            f"{data.std():>20.4f}"
            f"{(data == data.min()).sum():>20d}"
            f"{(data == data.max()).sum():>20d}"
        )

    return atoms_modi


def get_convex_hull(
    atoms, energy_name="energy", **kwargs
) -> tuple[np.ndarray, np.ndarray]:
    """
    Calculate the simple linear (E,V) convex hull.

    Parameters
    ----------
    atoms: list
        List of atoms objects.
    energy_name: str
        Name of the energy key in atoms.info (typically a DFT energy).

    Returns
    -------
    tuple
        A tuple containing two elements:
        - lower_half_hull_points: list of points (volume, energy) in the convex hull (lower half only).
        - p: list of all points for testing purposes.

    """
    points_list = []
    failed_count = 0

    for atom in atoms:
        if atom.info["config_type"] in ["IsolatedAtom", "dimer"]:
            continue
        try:
            volume_per_atom = atom.get_volume() / len(atom)
            energy_per_atom = atom.info[energy_name] / len(atom)
            points_list.append((volume_per_atom, energy_per_atom))
        except KeyError:
            failed_count += 1

    if failed_count > 0:
        raise ValueError(
            f"Convex hull failed to include {failed_count}/{len(atoms)} structures"
        )

    points = np.array(points_list)
    points = points.T[:, np.argsort(points.T[0])].T  # sort by volume axis

    hull = ConvexHull(points)  # generate full convex hull
    hull_points = points[hull.vertices]

    min_x_index = np.argmin(hull_points[:, 0])
    max_x_index = np.argmax(hull_points[:, 0])

    lower_half_hull = []
    i = min_x_index

    while True:
        lower_half_hull.append(hull.vertices[i])
        i = (i + 1) % len(hull.vertices)
        if i == max_x_index:
            lower_half_hull.append(hull.vertices[i])
            break

    lower_half_hull_points = points[lower_half_hull]

    lower_half_hull_points = lower_half_hull_points[
        lower_half_hull_points[:, 1] <= np.max(lower_half_hull_points[:, 1])
    ]

    return lower_half_hull_points, points


def get_e_distance_to_hull(
    hull: np.array, atoms, energy_name="energy", **kwargs
) -> float:
    """
    Calculate the distance of a structure to the linear convex hull in energy.

    Parameters
    ----------
    hull: (np.array)
        Points in the convex hull
    atoms: (Atoms)
        Structure to calculate distance to hull
    energy_name: (str)
        Name of energy key in atoms.info (typically a DFT energy)

    """
    volume = atoms.get_volume() / len(atoms)
    energy = atoms.info[energy_name] / len(atoms)
    tp = np.array([volume, energy])
    hull_ps = hull.points if isinstance(hull, ConvexHull) else hull

    if any(
        np.isclose(hull_ps[:], tp).all(1)
    ):  # if the point is on the hull, return 0.0
        return 0.0

    nearest = np.searchsorted(hull_ps.T[0], tp, side="right")[
        0
    ]  # find the nearest convex hull point

    return (
        energy
        - get_intersect(
            tp,  # get intersection of the vertical line (energy axis)
            tp + np.array([0, 1]),  # and the line between the nearest hull points
            hull_ps[(nearest - 1) % len(hull_ps.T[0])],
            hull_ps[nearest % len(hull_ps.T[0])],
        )[1]
    )


def get_intersect(a1, a2, b1, b2) -> tuple[float, float] | tuple:
    """
    Return the point of intersection of the lines passing through a2,a1 and b2,b1.

    a1: [x, y]
        A point on the first line
    a2: [x, y]
        Another point on the first line
    b1: [x, y]
        A point on the second line
    b2: [x, y]
        Another point on the second line

    """
    s = np.vstack([a1, a2, b1, b2])  # s for stacked
    h = np.hstack((s, np.ones((4, 1))))  # h for homogeneous
    l1 = np.cross(h[0], h[1])  # get first line
    l2 = np.cross(h[2], h[3])  # get second line
    x, y, z = np.cross(l1, l2)  # point of intersection
    if z == 0:  # lines are parallel
        return (float("inf"), float("inf"))
    return x / z, y / z


def get_mole_frac(atoms, element_order=None) -> float | int:
    """
    Calculate the mole-fraction of a structure.

    Parameters
    ----------
    atoms: (Atoms)
        Structure to calculate mole-fraction of
    element_order: (list)
        List of atomic numbers in order of choice (e.g. [42, 16] for MoS2)

    Returns
    -------
    (x2, x3...): (array of float)
        Reduced mole-fraction of structure - first element n = 1-sum(others)

    """
    element, cts = np.unique(atoms.get_atomic_numbers(), return_counts=True)

    if element_order is None and len(element) < 3:  # compatibility with old version
        x = cts[1] / sum(cts) if len(element) == 2 else 1

    else:  # new version, requires element_order, recommended for all new calculations
        if element_order is None:
            element_order = element  # use default order
        not_in = [i for i in element_order if i not in element]
        for i in not_in:
            element = np.insert(element, -1, i)
            cts = np.insert(cts, -1, 0)

        cts = np.array(
            [cts[np.argwhere(element == i).squeeze()] for i in element_order]
        )
        element = np.array(
            [element[np.argwhere(element == i).squeeze()] for i in element_order]
        )

        x = cts[1:] / sum(cts)

    return x


def label_stoichiometry_volume(
    atoms_list: list[Atoms],
    isolated_atom_energies: dict,
    energy_name: str,
    element_order: list | None = None,
) -> np.ndarray:
    """
    Calculate the stoichiometry, energy, and volume coordinates for forming the convex hull.

    Parameters
    ----------
    atoms_list: (list[Atoms])
        List of atoms objects
    isolated_atom_energies: (dict)
        Dictionary of isolated atom energies {atomic_number: energy}
    energy_name: (str)
        Name of energy key in atoms.info (typically a DFT energy)
    element_order: (list | None)
        List of atomic numbers in order of choice (e.g. [42, 16] for MoS2)

    """
    isolated_atom_energies = {
        ast.literal_eval(k) if isinstance(k, str) else k: v
        for k, v in isolated_atom_energies.items()
    }
    points_list = []
    for atom in atoms_list:
        try:
            volume = atom.get_volume() / len(atom)
            # make energy relative to isolated atoms
            energy = (
                atom.info[energy_name]
                - sum([isolated_atom_energies[j] for j in atom.get_atomic_numbers()])
            ) / len(atom)
            mole_frac = get_mole_frac(atom, element_order=element_order)
            points_list.append(np.hstack((mole_frac, volume, energy)))
        except KeyError:
            traceback.print_exc()
    points = np.array(points_list)
    return points.T[:, np.argsort(points.T[0])].T


def point_in_triangle_2D(p1, p2, p3, pn) -> bool:
    """
    Check if a point is inside a triangle in 2D.

    Parameters
    ----------
    p1: (tuple)
        Coordinates of first point
    p2: (tuple)
        Coordinates of second point
    p3: (tuple)
        Coordinates of third point
    pn: (tuple)
        Coordinates of point to check

    """
    ep = 1e-4
    x1, y1 = p1
    x2, y2 = p2
    x3, y3 = p3
    x, y = pn

    denominator = (y2 - y3) * (x1 - x3) + (x3 - x2) * (y1 - y3)
    a = ((y2 - y3) * (x - x3) + (x3 - x2) * (y - y3)) / denominator
    b = ((y3 - y1) * (x - x3) + (x1 - x3) * (y - y3)) / denominator
    c = 1 - a - b

    return (
        0 - ep <= a
        and a <= 1 + ep
        and 0 - ep <= b
        and b <= 1 + ep
        and 0 - ep <= c
        and c <= 1 + ep
    )


def point_in_triangle_nd(pn, *preg) -> bool:
    """
    Check if a point is inside a region of hyperplanes in N dimensions.

    Make a little convex hull in N-1 D and check this.

    Parameters
    ----------
    pn:
        Point to check (in ND)
    *preg:
        List of points defining (in ND) to check against

    """
    hull = Delaunay(preg)
    return hull.find_simplex(pn) >= 0


def calculate_hull_3d(points_3d) -> ConvexHull:
    """
    Calculate the convex hull in 3D.

    Parameters
    ----------
    points_3d:
        point in 3D

    Returns
    -------
    convex hull in 3D.

    """
    p0 = np.array(
        [
            (points_3d[:, i].max() - points_3d[:, i].min()) / 2 + points_3d[:, i].min()
            for i in range(2)
        ]
        + [-1e6]
    )  # test point to get the visible facets from below
    pn = np.vstack((p0, points_3d))

    hull = ConvexHull(pn, qhull_options="QG0")
    hull.remove_dim = []

    return hull


def calculate_hull_nd(points_nd) -> ConvexHull:
    """
    Calculate the convex hull in ND (N>=3).

    Parameters
    ----------
    points_nd:
        Point in ND.

    Returns
    -------
    Convex hull in ND.

    """
    p0 = np.array(
        [
            (points_nd[:, i].max() - points_nd[:, i].min()) / 2 + points_nd[:, i].min()
            for i in range(points_nd.shape[1] - 1)
        ]
        + [-1e6]
    )  # test point to get the visible facets from below
    pn = np.vstack((p0, points_nd))
    remove_dim = []

    for i in range(points_nd.shape[1]):
        if np.all(points_nd.T[i, 0] == points_nd.T[i, :]):
            pn = np.delete(pn, i, axis=1)
            print(f"Convex hull lower dimensional - removing dimension {i}")
            remove_dim.append(i)

    hull = ConvexHull(pn, qhull_options="QG0")
    print("done calculating hull")
    hull.remove_dim = remove_dim

    return hull


def get_e_distance_to_hull_3d(
    hull, atoms, isolated_atom_energies=None, energy_name="energy", element_order=None
) -> float:
    """
    Calculate the energy distance to the convex hull in 3D.

    Parameters
    ----------
    hull:
        Convex hull.
    atoms: (ase.Atoms)
        Structure to calculate mole-fraction of
    isolated_atom_energies: (dict)
        Dictionary of isolated atom energies
    energy_name: (str)
        Name of energy key in atoms.info (typically a DFT energy)
    element_order: (list)
        List of atomic numbers in order of choice (e.g. [42, 16] for MoS2)

    """
    isolated_atom_energies = {
        ast.literal_eval(k) if isinstance(k, str) else k: v
        for k, v in isolated_atom_energies.items()
    }
    mole_frac = get_mole_frac(atoms, element_order=element_order)
    energy = (
        atoms.info[energy_name]
        - sum([isolated_atom_energies[j] for j in atoms.get_atomic_numbers()])
    ) / len(atoms)
    volume = atoms.get_volume() / len(atoms)

    sp = np.hstack([mole_frac, volume, energy])
    for i in hull.remove_dim:
        sp = np.delete(sp, i)

    if len(sp[:-1]) == 1:
        # print('doing convexhull analysis in 1D')
        return get_e_distance_to_hull(hull, atoms, energy_name=energy_name)

    for _ct, visible_facet in enumerate(hull.simplices[hull.good]):
        if point_in_triangle_nd(sp[:-1], *hull.points[visible_facet][:, :-1]):
            n_3 = hull.points[visible_facet]
            energy = sp[-1]

            norm = np.cross(n_3[2] - n_3[0], n_3[1] - n_3[0])
            plane_norm = norm / np.linalg.norm(norm)  # plane normal
            plane_constant = np.dot(plane_norm, n_3[0])  # plane constant

            return (
                energy
                - (plane_constant - plane_norm[0] * sp[0] - plane_norm[1] * sp[1])
                / plane_norm[2]
            )

    print("Failed to find distance to hull")
    return 1e6


def piecewise_linear(x, vals) -> np.ndarray:
    """
    Piecewise linear.

    Parameters
    ----------
    x:
        The x value.
    vals:
        The values

    """
    i = np.searchsorted([v[0] for v in vals], x)
    f0 = (vals[i][0] - x) / (vals[i][0] - vals[i - 1][0])
    return f0 * np.array(vals[i - 1][1]) + (1.0 - f0) * np.array(vals[i][1])



"""General fitting jobs using several MLIPs available."""

from __future__ import annotations

from pathlib import Path

from jobflow import job

from autoplex.fitting.common.utils import (
    check_convergence,
    gap_fitting,
    jace_fitting,
    m3gnet_fitting,
    mace_fitting,
    nequip_fitting,
)

current_dir = Path(__file__).absolute().parent
GAP_DEFAULTS_FILE_PATH = current_dir / "mlip-phonon-defaults.json"


@job
def machine_learning_fit(
    database_dir: str | Path,
    species_list: list,
    isolated_atom_energies: dict | None = None,
    num_processes_fit: int = 32,
    auto_delta: bool = True,
    glue_xml: bool = False,
    glue_file_path: str = "glue.xml",
    mlip_type: str | None = None,
    ref_energy_name: str = "REF_energy",
    ref_force_name: str = "REF_forces",
    ref_virial_name: str = "REF_virial",
    use_defaults: bool = True,
    device: str = "cuda",
    hyperpara_opt: bool = False,
    **fit_kwargs,
):
    """
    Job for fitting potential(s).

    Parameters
    ----------
    database_dir: Str | Path
        Path to the directory containing the database.
    species_list: list
        List of element names (strings) involved in the training dataset
    isolated_atom_energies: dict
        Dictionary of isolated atoms energies.
    num_processes_fit: int
        Number of processes for fitting.
    auto_delta: bool
        Automatically determine delta for 2b, 3b and soap terms.
    glue_xml: bool
        Use the glue.xml core potential instead of fitting 2b terms.
    glue_file_path: str
        Name of the glue.xml file path.
    mlip_type: str
        Choose one specific MLIP type to be fitted:
        'GAP' | 'J-ACE' | 'NEQUIP' | 'M3GNET' | 'MACE'
    ref_energy_name: str
        Reference energy name.
    ref_force_name: str
        Reference force name.
    ref_virial_name: str
        Reference virial name.
    use_defaults: bool
        If True, use default fitting parameters
    device: str
        Device to be used for model fitting, either "cpu" or "cuda".
    hyperpara_opt: bool
        Perform hyperparameter optimization using XPOT
        (XPOT: https://pubs.aip.org/aip/jcp/article/159/2/024803/2901815)
    fit_kwargs: dict
        Additional keyword arguments for MLIP fitting.
    """
    if isinstance(database_dir, str):  # data_prep_job.output is returned as string
        database_dir = Path(database_dir)

    train_files = [
        "train.extxyz",
        "train_wo_sigma.extxyz",
        "train_phonon.extxyz",
        "train_rand_struc.extxyz",
    ]
    test_files = [
        "test.extxyz",
        "test.extxyz",
        "test_phonon.extxyz",
        "test_rand_struc.extxyz",
    ]

    if mlip_type == "GAP":
        for train_name, test_name in zip(train_files, test_files):
            if (database_dir / train_name).exists() and (
                database_dir / test_name
            ).exists():
                train_test_error = gap_fitting(
                    db_dir=database_dir,
                    species_list=species_list,
                    num_processes_fit=num_processes_fit,
                    auto_delta=auto_delta,
                    glue_xml=glue_xml,
                    glue_file_path=glue_file_path,
                    ref_energy_name=ref_energy_name,
                    ref_force_name=ref_force_name,
                    ref_virial_name=ref_virial_name,
                    train_name=train_name,
                    test_name=test_name,
                    fit_kwargs=fit_kwargs,
                )

    elif mlip_type == "J-ACE":
        train_test_error = jace_fitting(
            db_dir=database_dir,
            isolated_atom_energies=isolated_atom_energies,
            ref_energy_name=ref_energy_name,
            ref_force_name=ref_force_name,
            ref_virial_name=ref_virial_name,
            num_processes_fit=num_processes_fit,
            fit_kwargs=fit_kwargs,
        )

    elif mlip_type == "NEQUIP":
        train_test_error = nequip_fitting(
            db_dir=database_dir,
            isolated_atom_energies=isolated_atom_energies,
            ref_energy_name=ref_energy_name,
            ref_force_name=ref_force_name,
            ref_virial_name=ref_virial_name,
            fit_kwargs=fit_kwargs,
            device=device,
        )

    elif mlip_type == "M3GNET":
        train_test_error = m3gnet_fitting(
            db_dir=database_dir,
            ref_energy_name=ref_energy_name,
            ref_force_name=ref_force_name,
            ref_virial_name=ref_virial_name,
            fit_kwargs=fit_kwargs,
            device=device,
        )

    elif mlip_type == "MACE":
        train_test_error = mace_fitting(
            db_dir=database_dir,
            ref_energy_name=ref_energy_name,
            ref_force_name=ref_force_name,
            ref_virial_name=ref_virial_name,
            use_defaults=use_defaults,
            device=device,
            fit_kwargs=fit_kwargs,
        )

    check_conv = check_convergence(train_test_error["test_error"])

    return {
        "mlip_path": train_test_error["mlip_path"],
        "train_error": train_test_error["train_error"],
        "test_error": train_test_error["test_error"],
        "convergence": check_conv,
    }



import numpy as np
from pymatgen.core.structure import Structure
from pymatgen.phonon.bandstructure import PhononBandStructureSymmLine
from pymatgen.phonon.dos import PhononDos

from atomate2.common.schemas.phonons import (
    PhononBSDOSDoc,
    PhononComputationalSettings,
    PhononJobDirs,
    PhononUUIDs,
    ThermalDisplacementData,
)
from atomate2.vasp.flows.phonons import PhononMaker


def test_phonon_wf_only_displacements3(mock_vasp, clean_dir):
    from jobflow import run_locally

    structure = Structure(
        lattice=[[0, 2.73, 2.73], [2.73, 0, 2.73], [2.73, 2.73, 0]],
        species=["Si", "Si"],
        coords=[[0, 0, 0], [0.25, 0.25, 0.25]],
    )

    # mapping from job name to directory containing test files
    ref_paths = {
        "phonon static 1/1": "Si_phonons_2/phonon_static_1_1",
        "static": "Si_phonons_2/static",
    }

    # settings passed to fake_run_vasp; adjust these to check for certain INCAR settings
    fake_run_vasp_kwargs = {
        "phonon static 1/1": {"incar_settings": ["NSW", "ISMEAR"]},
        "static": {"incar_settings": ["NSW", "ISMEAR"]},
    }

    # automatically use fake VASP and write POTCAR.spec during the test
    mock_vasp(ref_paths, fake_run_vasp_kwargs)

    # !!! Generate job
    job = PhononMaker(
        min_length=3.0,
        bulk_relax_maker=None,
        born_maker=None,
        use_symmetrized_structure="conventional",
        create_thermal_displacements=False,
        store_force_constants=False,
        prefer_90_degrees=False,
        generate_frequencies_eigenvectors_kwargs={"tstep": 100},
        get_supercell_size_kwargs={"step_size": 0.5}
    ).make(structure)

    # run the flow or job and ensure that it finished running successfully
    responses = run_locally(job, create_folders=True, ensure_success=True)

    # !!! validation on the outputs
    assert isinstance(responses[job.jobs[-1].uuid][1].output, PhononBSDOSDoc)

    assert np.allclose(
        responses[job.jobs[-1].uuid][1].output.free_energies,
        [
            5774.603553771463,
            5616.334060911681,
            4724.766198084037,
            3044.208072582665,
            696.3373193497828,
        ],
    )

    assert isinstance(
        responses[job.jobs[-1].uuid][1].output.phonon_bandstructure,
        PhononBandStructureSymmLine,
    )
    assert isinstance(responses[job.jobs[-1].uuid][1].output.phonon_dos, PhononDos)
    assert responses[job.jobs[-1].uuid][1].output.thermal_displacement_data is None
    assert isinstance(responses[job.jobs[-1].uuid][1].output.structure, Structure)
    assert np.allclose(
        responses[job.jobs[-1].uuid][1].output.temperatures, [0, 100, 200, 300, 400]
    )
    assert np.allclose(
        responses[job.jobs[-1].uuid][1].output.has_imaginary_modes, False
    )
    assert responses[job.jobs[-1].uuid][1].output.force_constants is None
    assert isinstance(responses[job.jobs[-1].uuid][1].output.jobdirs, PhononJobDirs)
    assert isinstance(responses[job.jobs[-1].uuid][1].output.uuids, PhononUUIDs)
    assert np.isclose(
        responses[job.jobs[-1].uuid][1].output.total_dft_energy, -5.74555232
    )
    assert responses[job.jobs[-1].uuid][1].output.born is None
    assert responses[job.jobs[-1].uuid][1].output.epsilon_static is None
    assert np.allclose(
        responses[job.jobs[-1].uuid][1].output.supercell_matrix,
        [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
    )
    assert np.allclose(
        responses[job.jobs[-1].uuid][1].output.primitive_matrix,
        (
            (0, 0.5000000000000001, 0.5000000000000001),
            (0.5000000000000001, 0.0, 0.5000000000000001),
            (0.5000000000000001, 0.5000000000000001, 0.0),
        ),
    )
    assert responses[job.jobs[-1].uuid][1].output.code == "vasp"
    assert isinstance(
        responses[job.jobs[-1].uuid][1].output.phonopy_settings,
        PhononComputationalSettings,
    )
    assert responses[job.jobs[-1].uuid][1].output.phonopy_settings.npoints_band == 101
    assert (
        responses[job.jobs[-1].uuid][1].output.phonopy_settings.kpath_scheme
        == "seekpath"
    )
    assert (
        responses[job.jobs[-1].uuid][1].output.phonopy_settings.kpoint_density_dos
        == 7000
    )
    assert np.allclose(
        responses[job.jobs[-1].uuid][1].output.entropies,
        [
            0.0,
            4.786689009990746,
            13.02544271435008,
            20.360935069065423,
            26.39830736008501,
        ],
    )
    assert np.allclose(
        responses[job.jobs[-1].uuid][1].output.heat_capacities,
        [
            0.0,
            8.047577577838897,
            15.97117761314484,
            19.97051059716143,
            21.87494655884403,
        ],
    )

    assert np.allclose(
        responses[job.jobs[-1].uuid][1].output.internal_energies,
        [
            5774.603553771463,
            6095.002960937394,
            7329.854739783488,
            9152.488591840654,
            11255.660261586278,
        ],
    )


def test_cleanup_auxiliary_files(test_dir, clean_dir):
    import os
    import shutil

    # for some reason, not all the auxiliary files are removed at the end
    os.chdir(test_dir / "data" / "ref_data")
    for file_name in os.listdir(os.getcwd()):
        if file_name not in ['train_Si.extxyz', 'test_Si.extxyz', 'quip_train_Si.extxyz', 'quip_test_Si.extxyz']:
            file_path = os.path.join(os.getcwd(), file_name)
            try:
                # Check if it's a file or directory
                if os.path.isfile(file_path) or os.path.islink(file_path):
                    os.remove(file_path)  # Remove the file
                elif os.path.isdir(file_path):
                    shutil.rmtree(file_path)  # Remove the directory and its contents
            except Exception as e:
                print(f'Failed to delete {file_path}. Reason: {e}')



"""
AutoPLEX -- Automated machine-learned Potential Landscape explorer -- flows.

This package is designed to perform a workflow-based
automated machine-learned interatomic potentials (MLIP) fit for
DFT-labelled data.

"""



"""RSS Jobs include the generation of the initial potential model as well as iterative RSS exploration."""

from __future__ import annotations

import logging

from jobflow import Flow, Response, job

from autoplex.data.common.flows import DFTStaticLabelling
from autoplex.data.common.jobs import (
    collect_dft_data,
    preprocess_data,
    sample_data,
)
from autoplex.data.rss.flows import BuildMultiRandomizedStructure
from autoplex.data.rss.jobs import do_rss_multi_node
from autoplex.fitting.common.flows import MLIPFitMaker

__all__ = ["initial_rss", "do_rss_iterations"]

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)


@job
def initial_rss(
    tag: str,
    generated_struct_numbers: list[int],
    num_of_initial_selected_structs: list[int] | None = None,
    buildcell_options: list[dict] | None = None,
    fragment_file: str | None = None,
    fragment_numbers: list[str] | None = None,
    num_processes_buildcell: int = 1,
    initial_selection_enabled: bool = False,
    bcur_params: dict | None = None,
    random_seed: int | None = None,
    include_isolated_atom: bool = False,
    isolatedatom_box: list[float] | None = None,
    e0_spin: bool = False,
    include_dimer: bool = False,
    dimer_box: list[float] | None = None,
    dimer_range: list | None = None,
    dimer_num: int = 21,
    custom_incar: dict | None = None,
    custom_potcar: dict | None = None,
    config_type: str | None = None,
    vasp_ref_file: str = "vasp_ref.extxyz",
    rss_group: str = "initial",
    test_ratio: float = 0.1,
    regularization: bool = False,
    retain_existing_sigma: bool = False,
    scheme: str | None = None,
    reg_minmax: list[tuple] | None = None,
    distillation: bool = False,
    force_max: float | None = None,
    force_label: str = "REF_forces",
    pre_database_dir: str | None = None,
    mlip_type: str = "GAP",
    ref_energy_name: str = "REF_energy",
    ref_force_name: str = "REF_forces",
    ref_virial_name: str = "REF_virial",
    auto_delta: bool = False,
    num_processes_fit: int = 1,
    device_for_fitting: str = "cpu",
    **fit_kwargs,
):
    """
    Run initial Random Structure Searching (RSS) workflow from scratch.

    Parameters
    ----------
    tag: str
        Tag of systems. It can also be used for setting up elements and stoichiometry.
        For example, 'SiO2' will generate structures with a 2:1 ratio of Si to O.
    generated_struct_numbers: list[int]
        Expected number of generated randomized unit cells.
    num_of_initial_selected_structs: list[int] | None
        Number of structures to be sampled. Default is None.
    buildcell_options: list[dict] | None
        Customized parameters for buildcell. Default is None.
    fragment_file: Atoms | list[Atoms] | None
        Fragment(s) for random structures, e.g. molecules, to be placed indivudally intact.
        atoms.arrays should have a 'fragment_id' key with unique identifiers for each fragment if in same Atoms.
        atoms.cell must be defined (e.g. Atoms.cell = np.eye(3)*20).
    fragment_numbers: list[str] | None
        Numbers of each fragment to be included in the random structures. Defaults to 1 for all specified.
    num_processes_buildcell: int
        Number of processes to use for parallel computation during buildcell generation. Default is 1.
    initial_selection_enabled: bool
        If true, sample structures using CUR. Default is False.
    bcur_params: dict | None
        Parameters for Boltzmann CUR selection. Default is None.
    random_seed: int | None
        A seed to ensure reproducibility of CUR selection. Default is None.
    include_isolated_atom: bool
        If true, perform single-point calculations for isolated atoms. Default is False.
    isolatedatom_box: list[float] | None
        List of the lattice constants for an isolated atom configuration. Default is None.
    e0_spin: bool
        If true, include spin polarization in isolated atom and dimer calculations. Default is False.
    include_dimer: bool
        If true, perform single-point calculations for dimers. Default is False.
    dimer_box: list[float] | None
        The lattice constants of a dimer box. Default is None.
    dimer_range: list[float] | None
        Range of distances for dimer calculations. Default is None.
    dimer_num: int
        Number of different distances to consider for dimer calculations. Default is 21.
    custom_incar: dict | None
        Dictionary of custom VASP input parameters. If provided, will update the
        default parameters. Default is None.
    custom_potcar: dict | None
        Dictionary of POTCAR settings to update. Keys are element symbols, values are the desired POTCAR labels.
        Default is None.
    config_type: str | None
        Configuration type for the VASP calculations. Default is None.
    vasp_ref_file: str
        Reference file for VASP data. Default is 'vasp_ref.extxyz'.
    rss_group: str
        Group name for GAP RSS. Default is 'initial'.
    test_ratio: float
        The proportion of the test set after splitting the data.
        If None, no splitting will be performed. Default is 0.1.
    regularization: bool
        If true, apply regularization. This only works for GAP. Default is False.
    retain_existing_sigma: bool
        Whether to keep the current sigma values for specific configuration types.
        If set to True, existing sigma values for specific configurations will remain unchanged.
    scheme: str | None
        Scheme to use for regularization. Default is None.
    reg_minmax: list[tuple] | None
        A list of tuples representing the minimum and maximum values for regularization.
    distillation: bool
        If true, apply data distillation. Default is False.
    force_max: float | None
        Maximum force value to exclude structures. Default is None.
    force_label: str
        The label of force values to use for distillation. Default is 'REF_forces'.
    pre_database_dir: str | None
        Directory where the previous database was saved. Default is None.
    mlip_type: str
        Choose one specific MLIP type to be fitted: 'GAP' | 'J-ACE' | 'NEQUIP' | 'M3GNET' | 'MACE'.
        Default is 'GAP'.
    ref_energy_name: str
        Reference energy name. Default is 'REF_energy'.
    ref_force_name: str
        Reference force name. Default is 'REF_forces'.
    ref_virial_name: str
        Reference virial name. Default is 'REF_virial'.
    auto_delta: bool
        If true, apply automatic determination of delta for GAP terms. Default is False.
    num_processes_fit: int
        Number of processes used for fitting. Default is 1.
    device_for_fitting: str
        Device to be used for model fitting, either "cpu" or "cuda".
    fit_kwargs:
        Additional keyword arguments for the MLIP fitting process.

    Returns
    -------
    dict
        A dictionary whose keys contains:
            test_error: float
                The test error of the fitted MLIP.
            pre_database_dir: str
                The directory of the preprocessed database.
            mlip_path: str
                The path to the fitted MLIP.
            isolated_atom_energies: dict
                The isolated energy values.
            current_iter: int
                The current iteration index, set to 0.
    """
    if isolatedatom_box is None:
        isolatedatom_box = [20.0, 20.0, 20.0]
    if dimer_box is None:
        dimer_box = [20.0, 20.0, 20.0]

    do_randomized_structure_generation = BuildMultiRandomizedStructure(
        generated_struct_numbers=generated_struct_numbers,
        buildcell_options=buildcell_options,
        fragment_file=fragment_file,
        fragment_numbers=fragment_numbers,
        selected_struct_numbers=num_of_initial_selected_structs,
        tag=tag,
        num_processes=num_processes_buildcell,
        initial_selection_enabled=initial_selection_enabled,
        bcur_params=bcur_params,
        random_seed=random_seed,
    ).make()
    do_dft_static = DFTStaticLabelling(
        e0_spin=e0_spin,
        isolatedatom_box=isolatedatom_box,
        isolated_atom=include_isolated_atom,
        dimer=include_dimer,
        dimer_box=dimer_box,
        dimer_range=dimer_range,
        dimer_num=dimer_num,
        custom_incar=custom_incar,
        custom_potcar=custom_potcar,
    ).make(
        structures=do_randomized_structure_generation.output, config_type=config_type
    )
    do_data_collection = collect_dft_data(
        vasp_ref_file=vasp_ref_file, rss_group=rss_group, vasp_dirs=do_dft_static.output
    )
    do_data_preprocessing = preprocess_data(
        test_ratio=test_ratio,
        regularization=regularization,
        retain_existing_sigma=retain_existing_sigma,
        scheme=scheme,
        distillation=distillation,
        force_max=force_max,
        force_label=force_label,
        vasp_ref_dir=do_data_collection.output["vasp_ref_dir"],
        pre_database_dir=pre_database_dir,
        reg_minmax=reg_minmax,
        isolated_atom_energies=do_data_collection.output["isolated_atom_energies"],
    )
    do_mlip_fit = MLIPFitMaker(
        mlip_type=mlip_type,
        ref_energy_name=ref_energy_name,
        ref_force_name=ref_force_name,
        ref_virial_name=ref_virial_name,
    ).make(
        database_dir=do_data_preprocessing.output,
        isolated_atom_energies=do_data_collection.output["isolated_atom_energies"],
        num_processes_fit=num_processes_fit,
        apply_data_preprocessing=False,
        auto_delta=auto_delta,
        glue_xml=False,
        device=device_for_fitting,
        **fit_kwargs,
    )

    job_list = [
        do_randomized_structure_generation,
        do_dft_static,
        do_data_collection,
        do_data_preprocessing,
        do_mlip_fit,
    ]

    return Response(
        replace=Flow(job_list),
        output={
            "test_error": do_mlip_fit.output["test_error"],
            "pre_database_dir": do_data_preprocessing.output,
            "mlip_path": do_mlip_fit.output["mlip_path"],
            "isolated_atom_energies": do_data_collection.output[
                "isolated_atom_energies"
            ],
        },
    )


@job
def do_rss_iterations(
    input: dict,
    tag: str,
    generated_struct_numbers: list[int],
    num_of_initial_selected_structs: list[int] | None = None,
    buildcell_options: list[dict] | None = None,
    fragment_file: str | None = None,
    fragment_numbers: list[str] | None = None,
    num_processes_buildcell: int = 1,
    initial_selection_enabled: bool = False,
    rss_selection_method: str = None,
    num_of_rss_selected_structs: int = 100,
    bcur_params: dict | None = None,
    random_seed: int | None = None,
    include_isolated_atom: bool = False,
    isolatedatom_box: list[float] | None = None,
    e0_spin: bool = False,
    include_dimer: bool = False,
    dimer_box: list[float] | None = None,
    dimer_range: list | None = None,
    dimer_num: int = 21,
    custom_incar: dict | None = None,
    custom_potcar: dict | None = None,
    config_types: list[str] | None = None,
    vasp_ref_file: str = "vasp_ref.extxyz",
    rss_group: str = "rss",
    test_ratio: float = 0.1,
    regularization: bool = False,
    retain_existing_sigma: bool = False,
    scheme: str | None = None,
    reg_minmax: list[tuple] | None = None,
    distillation: bool = True,
    force_max: float = 200,
    force_label: str = "REF_forces",
    mlip_type: str = "GAP",
    ref_energy_name: str = "REF_energy",
    ref_force_name: str = "REF_forces",
    ref_virial_name: str = "REF_virial",
    auto_delta: bool = False,
    num_processes_fit: int = 1,
    device_for_fitting: str = "cpu",
    scalar_pressure_method: str = "exp",
    scalar_exp_pressure: float = 100,
    scalar_pressure_exponential_width: float = 0.2,
    scalar_pressure_low: float = 0,
    scalar_pressure_high: float = 50,
    max_steps: int = 200,
    force_tol: float = 0.05,
    stress_tol: float = 0.05,
    hookean_repul: bool = False,
    hookean_paras: dict[tuple[int, int], tuple[float, float]] | None = None,
    keep_symmetry: bool = False,
    write_traj: bool = True,
    num_processes_rss: int = 1,
    device_for_rss: str = "cpu",
    stop_criterion: float = 0.01,
    max_iteration_number: int = 5,
    num_groups: int = 1,
    initial_kt: float = 0.3,
    current_iter_index: int = 1,
    **fit_kwargs,
):
    """
    Perform iterative RSS to improve the accuracy of a MLIP.

    Each iteration involves generating new structures, sampling, running
    VASP calculations, collecting data, preprocessing data, and fitting a new MLIP.

    Parameters
    ----------
    input : dict
        A dictionary parameter used to pass specific input data required during the RSS iterations.
        The keys in this dictionary should be one of the following valid keys:
            test_error: float
                The test error of the fitted MLIP.
            pre_database_dir: str
                The directory of the preprocessed database.
            mlip_path: str
                The path to the fitted MLIP.
            isolated_atom_energies: dict
                The isolated energy values.
            current_iter: int
                The current iteration index.
            kt: float
                The value of kt.
    tag: str
        Tag of systems. It can also be used for setting up elements and stoichiometry.
        For example, 'SiO2' will generate structures with a 2:1 ratio of Si to O.
    generated_struct_numbers: list[int]
        Expected number of generated randomized unit cells.
    num_of_initial_selected_structs: list[int] | None
        Number of structures to be sampled. Default is None.
    buildcell_options: list[dict] | None
        Customized parameters for buildcell. Default is None.
    fragment_file: Atoms | list[Atoms] | None
        Fragment(s) for random structures, e.g. molecules, to be placed indivudally intact.
        atoms.arrays should have a 'fragment_id' key with unique identifiers for each fragment if in same Atoms.
        atoms.cell must be defined (e.g. Atoms.cell = np.eye(3)*20).
    fragment_numbers: list[str] | None
        Numbers of each fragment to be included in the random structures. Defaults to 1 for all specified.
    num_processes_buildcell: int
        Number of processes to use for parallel computation during buildcell generation. Default is 1.
    initial_selection_enabled: bool
        If true, sample structures using CUR. Default is False.
    rss_selection_method: str
        Method for selecting samples from the generated structures. Default is None.
    num_of_rss_selected_structs: int
        Number of structures to be selected.
    bcur_params: dict | None
        Parameters for Boltzmann CUR selection. Default is None.
    random_seed: int | None
        A seed to ensure reproducibility of CUR selection. Default is None.
    include_isolated_atom: bool
        If true, perform single-point calculations for isolated atoms. Default is False.
    isolatedatom_box: list[float] | None
        List of the lattice constants for an isolated atom configuration. Default is None.
    e0_spin: bool
        If true, include spin polarization in isolated atom and dimer calculations. Default is False.
    include_dimer: bool
        If true, perform single-point calculations for dimers only once. Default is False.
    dimer_box: list[float] | None
        The lattice constants of a dimer box. Default is None.
    dimer_range: list[float] | None
        Range of distances for dimer calculations. Default is None.
    dimer_num: int
        Number of different distances to consider for dimer calculations. Default is 21.
    custom_incar: dict | None
        Dictionary of custom VASP input parameters. If provided, will update the
        default parameters. Default is None.
    custom_potcar: dict | None
        Dictionary of POTCAR settings to update. Keys are element symbols, values are the desired POTCAR labels.
        Default is None.
    config_types: list[str] | None
        Configuration types for the VASP calculations. Default is None.
    vasp_ref_file: str
        Reference file for VASP data. Default is 'vasp_ref.extxyz'.
    rss_group: str
        Group name for GAP RSS. Default is 'rss'.
    test_ratio: float
        The proportion of the test set after splitting the data. Default is 0.1.
    regularization: bool
        If true, apply regularization. This only works for GAP. Default is False.
    retain_existing_sigma: bool
        Whether to keep the current sigma values for specific configuration types.
        If set to True, existing sigma values for specific configurations will remain unchanged.
    scheme: str | None
        Scheme to use for regularization. Default is None.
    reg_minmax: list[tuple] | None
        A list of tuples representing the minimum and maximum values for regularization.
    distillation: bool
        If true, apply data distillation. Default is True.
    force_max: float
        Maximum force value to exclude structures. Default is 200.
    force_label: str
        The label of force values to use for distillation. Default is 'REF_forces'.
    mlip_type: str
        Choose one specific MLIP type: 'GAP' | 'J-ACE' | 'NequIP' | 'M3GNet' | 'MACE'. Default is 'GAP'.
    ref_energy_name: str
        Reference energy name. Default is 'REF_energy'.
    ref_force_name: str
        Reference force name. Default is 'REF_forces'.
    ref_virial_name: str
        Reference virial name. Default is 'REF_virial'.
    auto_delta: bool
        If true, apply automatic determination of delta for GAP terms. Default is False.
    num_processes_fit: int
        Number of processes used for fitting. Default is 1.
    device_for_fitting: str
        Device to be used for model fitting, either "cpu" or "cuda".
    scalar_pressure_method: str
        Method for adding external pressures. Default is 'exp'.
    scalar_exp_pressure: float
        Scalar exponential pressure. Default is 100.
    scalar_pressure_exponential_width: float
        Width for scalar pressure exponential. Default is 0.2.
    scalar_pressure_low: float
        Low limit for scalar pressure. Default is 0.
    scalar_pressure_high: float
        High limit for scalar pressure. Default is 50.
    max_steps: int
        Maximum number of steps for relaxation. Default is 200.
    force_tol: float
        Force residual tolerance for relaxation. Default is 0.05.
    stress_tol: float
        Stress residual tolerance for relaxation. Default is 0.05.
    hookean_repul: bool
        If true, apply Hookean repulsion. Default is False.
    hookean_paras: dict[tuple[int, int], tuple[float, float]] | None
        Parameters for Hookean repulsion as a dictionary of tuples. Default is None.
    keep_symmetry: bool
        If true, preserve symmetry during relaxation. Default is False.
    write_traj: bool
        If true, write trajectory of RSS. Default is True.
    num_processes_rss: int
        Number of processes used for running RSS. Default is 1.
    device_for_rss: str
        Specify device to use "cuda" or "cpu" for running RSS. Default is "cpu".
    stop_criterion: float
        Convergence criterion for stopping RSS iterations. Default is 0.01.
    max_iteration_number: int
        Maximum number of RSS iterations to perform. Default is 5.
    num_groups: int
        Number of structure groups, used for assigning tasks across multiple nodes. Default is 1.
    initial_kt: float
        Initial temperature (in eV) for Boltzmann sampling. Default is 0.3.
    current_iter_index: int
        Index for the current RSS iteration. Default is 1.
    fit_kwargs:
        Additional keyword arguments for the MLIP fitting process.

    Returns
    -------
    dict
        A dictionary whose keys contains:
            test_error: float
                The test error of the fitted MLIP.
            pre_database_dir: str
                The directory of the preprocessed database.
            mlip_path: str
                The path to the fitted MLIP.
            isolated_atom_energies: dict
                The isolated energy values.
            current_iter: int
                The current iteration index.
            kt: float
                The temperature (in eV) for Boltzmann sampling.
    """
    test_error = input.get("test_error")
    current_iter = input.get("current_iter", current_iter_index)
    current_kt = input.get("kt", initial_kt)

    config_type = (
        (config_types[0] if current_kt > 0.1 else config_types[-1])
        if config_types
        else None
    )

    if isolatedatom_box is None:
        isolatedatom_box = [20.0, 20.0, 20.0]
    if dimer_box is None:
        dimer_box = [20.0, 20.0, 20.0]

    logging.info(
        f"The configuration type of structures generated in the current iteration will be {config_type}!"
    )

    if (
        test_error is not None
        and test_error > stop_criterion
        and current_iter is not None
        and current_iter < max_iteration_number
    ):
        logging.info(f"Current kt: {current_kt}")
        logging.info(f"Current iter index: {current_iter}")
        logging.info(f"The error of {current_iter}th iteration: {test_error}")

        if bcur_params is None:
            bcur_params = {}
        bcur_params["kt"] = current_kt

        do_randomized_structure_generation = BuildMultiRandomizedStructure(
            generated_struct_numbers=generated_struct_numbers,
            buildcell_options=buildcell_options,
            fragment_file=fragment_file,
            fragment_numbers=fragment_numbers,
            selected_struct_numbers=num_of_initial_selected_structs,
            tag=tag,
            num_processes=num_processes_buildcell,
            initial_selection_enabled=initial_selection_enabled,
            bcur_params=bcur_params,
            random_seed=random_seed,
        ).make()
        do_rss = do_rss_multi_node(
            mlip_type=mlip_type,
            iteration_index=f"{current_iter}th",
            mlip_path=input["mlip_path"],
            structure_paths=do_randomized_structure_generation.output,
            scalar_pressure_method=scalar_pressure_method,
            scalar_exp_pressure=scalar_exp_pressure,
            scalar_pressure_exponential_width=scalar_pressure_exponential_width,
            scalar_pressure_low=scalar_pressure_low,
            scalar_pressure_high=scalar_pressure_high,
            max_steps=max_steps,
            force_tol=force_tol,
            stress_tol=stress_tol,
            hookean_repul=hookean_repul,
            hookean_paras=hookean_paras,
            keep_symmetry=keep_symmetry,
            write_traj=write_traj,
            num_processes_rss=num_processes_rss,
            device=device_for_rss,
            num_groups=num_groups,
            config_type=config_type,
        )
        do_data_sampling = sample_data(
            selection_method=rss_selection_method,
            num_of_selection=num_of_rss_selected_structs,
            bcur_params=bcur_params,
            traj_path=do_rss.output,
            random_seed=random_seed,
            isolated_atom_energies=input["isolated_atom_energies"],
        )
        do_dft_static = DFTStaticLabelling(
            e0_spin=e0_spin,
            isolatedatom_box=isolatedatom_box,
            isolated_atom=include_isolated_atom,
            dimer=include_dimer,
            dimer_box=dimer_box,
            dimer_range=dimer_range,
            dimer_num=dimer_num,
            custom_incar=custom_incar,
            custom_potcar=custom_potcar,
        ).make(structures=do_data_sampling.output, config_type=config_type)
        do_data_collection = collect_dft_data(
            vasp_ref_file=vasp_ref_file,
            rss_group=rss_group,
            vasp_dirs=do_dft_static.output,
        )
        do_data_preprocessing = preprocess_data(
            test_ratio=test_ratio,
            regularization=regularization,
            retain_existing_sigma=retain_existing_sigma,
            scheme=scheme,
            distillation=distillation,
            force_max=force_max,
            force_label=force_label,
            vasp_ref_dir=do_data_collection.output["vasp_ref_dir"],
            pre_database_dir=input["pre_database_dir"],
            reg_minmax=reg_minmax,
            isolated_atom_energies=input["isolated_atom_energies"],
        )
        do_mlip_fit = MLIPFitMaker(
            mlip_type=mlip_type,
            ref_energy_name=ref_energy_name,
            ref_force_name=ref_force_name,
            ref_virial_name=ref_virial_name,
        ).make(
            database_dir=do_data_preprocessing.output,
            isolated_atom_energies=input["isolated_atom_energies"],
            num_processes_fit=num_processes_fit,
            apply_data_preprocessing=False,
            auto_delta=auto_delta,
            glue_xml=False,
            device=device_for_fitting,
            **fit_kwargs,
        )

        kt = current_kt - 0.1 if (current_kt - 0.1) > 0.1 else 0.1
        current_iter += 1
        if include_isolated_atom:
            include_isolated_atom = False
        if include_dimer:
            include_dimer = False

        do_iteration = do_rss_iterations(
            input={
                "test_error": do_mlip_fit.output["test_error"],
                "pre_database_dir": do_data_preprocessing.output,
                "mlip_path": do_mlip_fit.output["mlip_path"],
                "isolated_atom_energies": input["isolated_atom_energies"],
                "current_iter": current_iter,
                "kt": kt,
            },
            generated_struct_numbers=generated_struct_numbers,
            num_of_initial_selected_structs=num_of_initial_selected_structs,
            tag=tag,
            buildcell_options=buildcell_options,
            fragment_file=fragment_file,
            fragment_numbers=fragment_numbers,
            num_processes_buildcell=num_processes_buildcell,
            initial_selection_enabled=initial_selection_enabled,
            rss_selection_method=rss_selection_method,
            num_of_rss_selected_structs=num_of_rss_selected_structs,
            bcur_params=bcur_params,
            random_seed=random_seed,
            e0_spin=e0_spin,
            isolatedatom_box=isolatedatom_box,
            include_isolated_atom=include_isolated_atom,
            include_dimer=include_dimer,
            dimer_box=dimer_box,
            dimer_range=dimer_range,
            dimer_num=dimer_num,
            custom_incar=custom_incar,
            custom_potcar=custom_potcar,
            config_types=config_types,
            vasp_ref_file=vasp_ref_file,
            rss_group=rss_group,
            test_ratio=test_ratio,
            regularization=regularization,
            retain_existing_sigma=retain_existing_sigma,
            scheme=scheme,
            reg_minmax=reg_minmax,
            distillation=distillation,
            force_max=force_max,
            force_label=force_label,
            mlip_type=mlip_type,
            ref_energy_name=ref_energy_name,
            ref_force_name=ref_force_name,
            ref_virial_name=ref_virial_name,
            auto_delta=auto_delta,
            num_processes_fit=num_processes_fit,
            scalar_pressure_method=scalar_pressure_method,
            scalar_exp_pressure=scalar_exp_pressure,
            scalar_pressure_exponential_width=scalar_pressure_exponential_width,
            scalar_pressure_low=scalar_pressure_low,
            scalar_pressure_high=scalar_pressure_high,
            max_steps=max_steps,
            force_tol=force_tol,
            stress_tol=stress_tol,
            hookean_repul=hookean_repul,
            hookean_paras=hookean_paras,
            keep_symmetry=keep_symmetry,
            write_traj=write_traj,
            num_processes_rss=num_processes_rss,
            device_for_rss=device_for_rss,
            stop_criterion=stop_criterion,
            max_iteration_number=max_iteration_number,
            num_groups=num_groups,
            initial_kt=initial_kt,
            current_iter_index=current_iter_index,
            **fit_kwargs,
        )

        job_list = [
            do_randomized_structure_generation,
            do_rss,
            do_data_sampling,
            do_dft_static,
            do_data_collection,
            do_data_preprocessing,
            do_mlip_fit,
            do_iteration,
        ]

        return Response(detour=job_list, output=do_iteration.output)

    return input



"""Flows consisting of jobs to fit ML potentials."""

from __future__ import annotations

import os
import shutil
from dataclasses import dataclass
from pathlib import Path

import ase.io
from jobflow import Flow, Maker, job

from autoplex.fitting.common.jobs import machine_learning_fit
from autoplex.fitting.common.regularization import set_custom_sigma
from autoplex.fitting.common.utils import (
    get_list_of_vasp_calc_dirs,
    vaspoutput_2_extended_xyz,
    write_after_distillation_data_split,
)

__all__ = [
    "MLIPFitMaker",
    "DataPreprocessing",
]


@dataclass
class MLIPFitMaker(Maker):
    """
    Maker to fit ML potentials based on DFT labelled reference data.

    This Maker will filter the provided dataset in a data preprocessing step and then proceed
    with the MLIP fit (default is GAP).

    Parameters
    ----------
    name : str
        Name of the flows produced by this maker.
    mlip_type: str
        Choose one specific MLIP type to be fitted:
        'GAP' | 'J-ACE' | 'NEQUIP' | 'M3GNET' | 'MACE'
    hyperpara_opt: bool
        Perform hyperparameter optimization using XPOT
        (XPOT: https://pubs.aip.org/aip/jcp/article/159/2/024803/2901815)
    ref_energy_name : str
        Reference energy name.
    ref_force_name : str
        Reference force name.
    ref_virial_name : str
        Reference virial name.
    glue_file_path: str
        Name of the glue.xml file path.
    use_defaults: bool
        If true, uses default fit parameters
    """

    name: str = "MLpotentialFit"
    mlip_type: str = "GAP"
    hyperpara_opt: bool = False
    ref_energy_name: str = "REF_energy"
    ref_force_name: str = "REF_forces"
    ref_virial_name: str = "REF_virial"
    glue_file_path: str = "glue.xml"
    use_defaults: bool = True

    # TO DO: Combine parameters used only for gap into one category (as noted below),
    # otherwise it will be too specific.
    def make(
        self,
        fit_input: dict | None = None,  # This is specific to phonon workflow
        species_list: list | None = None,
        isolated_atom_energies: dict | None = None,
        split_ratio: float = 0.4,
        force_max: float = 40.0,
        regularization: bool = False,  # This is only used for GAP.
        distillation: bool = True,
        separated: bool = False,
        pre_xyz_files: list[str] | None = None,
        pre_database_dir: str | None = None,
        atomwise_regularization_parameter: float = 0.1,  # This is only used for GAP.
        force_min: float = 0.01,  # unit: eV Å-1
        atom_wise_regularization: bool = True,  # This is only used for GAP.
        auto_delta: bool = False,  # This is only used for GAP.
        glue_xml: bool = False,  # This is only used for GAP.
        num_processes_fit: int | None = None,
        apply_data_preprocessing: bool = True,
        database_dir: Path | str | None = None,
        device: str = "cpu",
        **fit_kwargs,
    ):
        """
        Make a flow for fitting MLIP models.

        Parameters
        ----------
        fit_input: dict
            Output from the CompletePhononDFTMLDataGenerationFlow process.
        species_list: list
            List of element names (strings) involved in the training dataset
        isolated_atom_energies: dict
            Dictionary of isolated atoms energies.
        split_ratio: float
            Ratio to divide the dataset into training and test sets.
            A value of 0.1 means 90% training data and 10% test data
        force_max: float
            Maximum allowed force in the dataset.
        regularization: bool
            For using sigma regularization.
        distillation: bool
            For using data distillation.
        separated: bool
            Repeat the fit for each data_type available in the (combined) database.
        pre_xyz_files: list[str] or None
            Names of the pre-database train xyz file and test xyz file.
        pre_database_dir: str or None
            The pre-database directory.
        atomwise_regularization_parameter: float
            Regularization value for the atom-wise force components.
        force_min: float
            Minimal force cutoff value for atom-wise regularization.
        atom_wise_regularization: bool
            For including atom-wise regularization.
        auto_delta: bool
            Automatically determine delta for 2b, 3b and soap terms.
        glue_xml: bool
            Use the glue.xml core potential instead of fitting 2b terms.
        num_processes_fit: int
            Number of processes for fitting.
        apply_data_preprocessing: bool
            Determine whether to preprocess the data.
        database_dir: Path | str
            Path to the directory containing the database.
        device: str
            Device to be used for model fitting, either "cpu" or "cuda".
        fit_kwargs: dict
            Additional keyword arguments for MLIP fitting.
        """
        if self.mlip_type not in ["GAP", "J-ACE", "NEQUIP", "M3GNET", "MACE"]:
            raise ValueError(
                "Please correct the MLIP name!"
                "The current version ONLY supports the following models: GAP, J-ACE, NEQUIP, M3GNET, and MACE."
            )

        if apply_data_preprocessing:
            jobs = []
            data_prep_job = DataPreprocessing(
                split_ratio=split_ratio,
                regularization=regularization,
                separated=separated,
                distillation=distillation,
                force_max=force_max,
            ).make(
                fit_input=fit_input,
                pre_xyz_files=pre_xyz_files,
                pre_database_dir=pre_database_dir,
                force_min=force_min,
                atomwise_regularization_parameter=atomwise_regularization_parameter,
                atom_wise_regularization=atom_wise_regularization,
            )
            jobs.append(data_prep_job)

            mlip_fit_job = machine_learning_fit(
                database_dir=data_prep_job.output,
                isolated_atom_energies=isolated_atom_energies,
                num_processes_fit=num_processes_fit,
                auto_delta=auto_delta,
                glue_xml=glue_xml,
                glue_file_path=self.glue_file_path,
                mlip_type=self.mlip_type,
                hyperpara_opt=self.hyperpara_opt,
                ref_energy_name=self.ref_energy_name,
                ref_force_name=self.ref_force_name,
                ref_virial_name=self.ref_virial_name,
                use_defaults=self.use_defaults,
                device=device,
                species_list=species_list,
                **fit_kwargs,
            )
            jobs.append(mlip_fit_job)

            return Flow(jobs=jobs, output=mlip_fit_job.output, name=self.name)
        # this will only run if train.extxyz and test.extxyz files are present in the database_dir

        if isinstance(database_dir, str):
            database_dir = Path(database_dir)

        mlip_fit_job = machine_learning_fit(
            database_dir=database_dir,
            isolated_atom_energies=isolated_atom_energies,
            num_processes_fit=num_processes_fit,
            auto_delta=auto_delta,
            glue_xml=glue_xml,
            glue_file_path=self.glue_file_path,
            mlip_type=self.mlip_type,
            hyperpara_opt=self.hyperpara_opt,
            ref_energy_name=self.ref_energy_name,
            ref_force_name=self.ref_force_name,
            ref_virial_name=self.ref_virial_name,
            device=device,
            species_list=species_list,
            **fit_kwargs,
        )

        return Flow(jobs=mlip_fit_job, output=mlip_fit_job.output, name=self.name)


@dataclass
class DataPreprocessing(Maker):
    """
    Data preprocessing of the provided dataset.

    Parameters
    ----------
    name : str
        Name of the flows produced by this maker.
    split_ratio: float
        Parameter to divide the training set and the test set.
        A value of 0.1 means that the ratio of the training set to the test set is 9:1
    regularization: bool
        For using sigma regularization.
    separated: bool
        Repeat the fit for each data_type available in the (combined) database.
    distillation: bool
        For using data distillation.
    force_max: float
        Maximally allowed force in the data set.

    """

    name: str = "data_preprocessing_for_fitting"
    split_ratio: float = 0.5
    regularization: bool = False
    separated: bool = False
    distillation: bool = False
    force_max: float = 40.0

    @job
    def make(
        self,
        fit_input: dict,
        pre_database_dir: str | None = None,
        pre_xyz_files: list[str] | None = None,
        atomwise_regularization_parameter: float = 0.1,
        force_min: float = 0.01,  # unit: eV Å-1
        atom_wise_regularization: bool = True,
    ):
        """
        Maker for data preprocessing.

        Parameters
        ----------
        fit_input:
            Mixed list of dictionary and lists of the fit input data.
        pre_database_dir: str or None
            The pre-database directory.
        pre_xyz_files: list[str] or None
            Names of the pre-database train xyz file and test xyz file labelled by VASP.
        atomwise_regularization_parameter: float
            Regularization value for the atom-wise force components.
        force_min: float
            Minimal force cutoff value for atom-wise regularization.
        atom_wise_regularization: bool
            If True, includes atom-wise regularization.

        """
        if pre_xyz_files is None:
            pre_xyz_files = ["train.extxyz", "test.extxyz"]

        list_of_vasp_calc_dirs = get_list_of_vasp_calc_dirs(flow_output=fit_input)

        config_types = [
            key
            for key, value in fit_input.items()
            for key2, value2 in value.items()
            if key2 != "phonon_data"
            for _ in value2[0]
        ]

        data_types = [
            key2
            for key, value in fit_input.items()
            for key2, value2 in value.items()
            if key2 != "phonon_data"
            for _ in value2[0]
        ]

        if pre_database_dir and os.path.exists(pre_database_dir):
            current_working_directory = os.getcwd()

            if len(pre_xyz_files) == 1:
                for file_name in pre_xyz_files:
                    source_file_path = os.path.join(pre_database_dir, file_name)
                    destination_file_path = os.path.join(
                        current_working_directory, "vasp_ref.extxyz"
                    )
                    shutil.copy(source_file_path, destination_file_path)
                    print(
                        f"File {file_name} has been copied to {destination_file_path}"
                    )

        vaspoutput_2_extended_xyz(
            path_to_vasp_static_calcs=list_of_vasp_calc_dirs,
            config_types=config_types,
            data_types=data_types,
            f_min=force_min,
            regularization=atomwise_regularization_parameter,
            atom_wise_regularization=atom_wise_regularization,
        )

        write_after_distillation_data_split(
            self.distillation, self.force_max, self.split_ratio
        )

        # Merging database
        if pre_database_dir and os.path.exists(pre_database_dir):
            if len(pre_xyz_files) == 2:
                files_new = ["train.extxyz", "test.extxyz"]
                for file_name, file_new in zip(pre_xyz_files, files_new):
                    with (
                        open(os.path.join(pre_database_dir, file_name)) as pre_xyz_file,
                        open(file_new, "a") as xyz_file,
                    ):
                        xyz_file.write(pre_xyz_file.read())
                    print(f"File {file_name} has been copied to {file_new}")

            elif len(pre_xyz_files) > 2:
                raise ValueError(
                    "Please provide a train and a test extxyz file (two files in total) for the pre_xyz_files."
                )
        if self.regularization:
            atoms = ase.io.read("train.extxyz", index=":")
            ase.io.write("train_wo_sigma.extxyz", atoms, format="extxyz")
            atoms_with_sigma = set_custom_sigma(
                atoms,
                reg_minmax=[(0.1, 1), (0.001, 0.1), (0.0316, 0.316), (0.0632, 0.632)],
            )
            ase.io.write("train.extxyz", atoms_with_sigma, format="extxyz")
        if self.separated:
            atoms_train = ase.io.read("train.extxyz", index=":")
            atoms_test = ase.io.read("test.extxyz", index=":")
            for dt in set(data_types):
                data_type = dt.rstrip("_dir")
                if data_type != "iso_atoms":
                    for atoms in atoms_train + atoms_test:
                        if atoms.info["data_type"] == "iso_atoms":
                            ase.io.write(
                                f"vasp_ref_{data_type}.extxyz",
                                atoms,
                                format="extxyz",
                                append=True,
                            )
                        if atoms.info["data_type"] == data_type:
                            ase.io.write(
                                f"vasp_ref_{data_type}.extxyz",
                                atoms,
                                format="extxyz",
                                append=True,
                            )

                    write_after_distillation_data_split(
                        distillation=self.distillation,
                        force_max=self.force_max,
                        split_ratio=self.split_ratio,
                        vasp_ref_name=f"vasp_ref_{data_type}.extxyz",
                        train_name=f"train_{data_type}.extxyz",
                        test_name=f"test_{data_type}.extxyz",
                    )

        return Path.cwd()



"""Utility functions for benchmarking jobs."""

from __future__ import annotations

from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import numpy as np

if TYPE_CHECKING:
    from matplotlib.figure import Figure
    from pymatgen.core.structure import Structure
    from pymatgen.phonon.bandstructure import PhononBandStructureSymmLine
from pymatgen.phonon.plotter import PhononBSPlotter


def compute_bandstructure_benchmark_metrics(
    ml_model: str,
    structure: Structure,
    mp_id: str,
    ml_phonon_bs: PhononBandStructureSymmLine,
    dft_phonon_bs: PhononBandStructureSymmLine,
    ml_imag_modes: bool,
    dft_imag_modes: bool,
    atomwise_regularization_parameter: float,
    soap_dict: dict,
    suffix: str,
    displacement: float = 0.01,
):
    """
    Compute phonon band-structure benchmark metrics and generate associated plots.

    Parameters
    ----------
    ml_model: str
        ML model to be used. Default is GAP.
    structure : .Structure
        A structure object.
    mp_id:
        Materials Project ID.
    ml_phonon_bs: PhononBandStructureSymmLine.
        ML generated pymatgen phonon band-structure object.
    dft_phonon_bs: PhononBandStructureSymmLine.
        DFT generated pymatgen phonon band-structure object.
    ml_imag_modes: bool
        Whether the ML-based phonon band structure shows imaginary modes.
    dft_imag_modes: bool
        Whether the DFT-based phonon band structure shows imaginary modes.
    displacement: float
        Displacement distance for phonons
    atomwise_regularization_parameter: float
        Regularization value for the atom-wise force components.
    suffix: str
        GAP potential file suffix.
    soap_dict: dict
        Dictionary containing SOAP parameters.


    Returns
    -------
    dict including
       Overall root mean squared error between DFT and ML phonon band-structure.
    """
    # compute overall root mean squared error
    overall_rmse = get_rmse(ml_bs=ml_phonon_bs, dft_bs=dft_phonon_bs)

    # saves rmse k-dependent plot
    file_name = f"{structure.composition.reduced_formula}_rmse_phonons.pdf"
    _ = rmse_qdep_plot(
        ml_bs=ml_phonon_bs,
        dft_bs=dft_phonon_bs,
        which_q_path=2,
        file_name=file_name,
        img_format="pdf",
    )

    # saves DFT and ML phonon band-structure overlay plot
    file_name = f"{structure.composition.reduced_formula}_band_comparison.pdf"
    _ = compare_plot(
        ml_model=ml_model,
        ml_bs=ml_phonon_bs,
        dft_bs=dft_phonon_bs,
        file_name=file_name,
    )
    return {
        "benchmark_phonon_rmse": overall_rmse,
        "dft_imaginary_modes": dft_imag_modes,
        "ml_imaginary_modes": ml_imag_modes,
        "ml_model": ml_model,
        "mp_id": mp_id,
        "structure": structure,
        "displacement": displacement,
        "atomwise_regularization_parameter": atomwise_regularization_parameter,
        "soap_dict": soap_dict,
        "suffix": suffix,
    }


def get_rmse(
    ml_bs: PhononBandStructureSymmLine,
    dft_bs: PhononBandStructureSymmLine,
    q_dependent_rmse: bool = False,
) -> float | list[float]:
    """
    Compute root mean squared error (rmse) between DFT and ML phonon band-structure.

    Parameters
    ----------
    ml_bs : PhononBandStructureSymmLine.
        ML generated pymatgen phonon band-structure object
    dft_bs : PhononBandStructureSymmLine.
        DFT generated pymatgen phonon band-structure object
    q_dependent_rmse : bool.
        If true, method returns k-dependent rmse between the band-structures

    Returns
    -------
    float or list[float]
      Root mean squared error between DFT and ML phonon band-structure
    """
    diff = np.sort(ml_bs.bands, axis=0) - np.sort(dft_bs.bands, axis=0)

    rmse = np.sqrt(np.mean(diff**2))

    if q_dependent_rmse:
        diff_here = np.transpose(diff)
        rmse = [np.sqrt(np.mean(diff_here[i] ** 2)) for i in range(len(diff_here))]

    return rmse


def rmse_qdep_plot(
    ml_bs: PhononBandStructureSymmLine,
    dft_bs: PhononBandStructureSymmLine,
    which_q_path=1,
    file_name="rms.pdf",
    img_format="pdf",
) -> plt:
    """
    Save q dependent root mean squared error plot between DFT and ML phonon band-structure.

    Parameters
    ----------
    ml_bs : PhononBandStructureSymmLine.
        ML generated pymatgen phonon band-structure object.
    dft_bs : PhononBandStructureSymmLine.
        DFT generated pymatgen phonon band-structure object.
    which_q_path : int.
        If set 1, use ML band-structure as reference, if 2, uses DFT band-structure as reference.
    file_name: str.
        Name of the saved plot
    img_format: str
        File extension of plot to be saved, default is pdf.

    Returns
    -------
    matplotlib.plt
        A matplotlib figure with q-dependent RMSE.
    """
    rmse_qp = get_rmse(ml_bs=ml_bs, dft_bs=dft_bs, q_dependent_rmse=True)
    if which_q_path == 1:
        plotter = PhononBSPlotter(bs=ml_bs)
    elif which_q_path == 2:
        plotter = PhononBSPlotter(bs=dft_bs)

    distances = []
    for element in plotter.bs_plot_data()["distances"]:
        distances.extend(element)

    plt.close("all")
    plt.plot(distances, rmse_qp)
    plt.xticks(
        ticks=plotter.bs_plot_data()["ticks"]["distance"],
        labels=plotter.bs_plot_data()["ticks"]["label"],
    )
    plt.xlabel("Wave vector")
    plt.ylabel("Phonons RMS (THz)")
    plt.savefig(file_name, format=img_format)

    return plt


def compare_plot(
    ml_model: str,
    ml_bs: PhononBandStructureSymmLine,
    dft_bs: PhononBandStructureSymmLine,
    file_name: str = "band_comparison.pdf",
) -> Figure:
    """
    Save DFT and ML phonon band-structure overlay plot for visual comparison.

    Parameters
    ----------
    ml_model: str
        ML model to be used. Default is GAP.
    ml_bs : PhononBandStructureSymmLine.
        ML generated pymatgen phonon band-structure object
    dft_bs : PhononBandStructureSymmLine.
        DFT generated pymatgen phonon band-structure object
    file_name: str.
        Name of the saved plot

    Returns
    -------
    matplotlib.plt
        A matplotlib figure with DFT and ML generated phonon band-structure overlay
    """
    plotter = PhononBSPlotter(bs=ml_bs)
    plotter2 = PhononBSPlotter(bs=dft_bs)
    new_plotter = plotter.plot_compare(
        other_plotter={"DFT": plotter2}, self_label=ml_model
    )

    new_plotter.figure.savefig(fname=file_name)

    return new_plotter.figure



"""RSS (random structure searching) flow for exploring and learning potential energy surfaces from scratch."""

from __future__ import annotations

import os
from dataclasses import dataclass
from pathlib import Path

from jobflow import Flow, Maker, Response, job
from ruamel.yaml import YAML

from autoplex.auto.rss.jobs import do_rss_iterations, initial_rss


@dataclass
class RssMaker(Maker):
    """
    Maker to set up and run RSS for exploring and learning potential-energy surfaces (from scratch).

    Parameters
    ----------
    name: str
        Name of the flow.
    path_to_default_config_parameters: Path | str | None
        Path to the default RSS configuration file 'rss_default_configuration.yaml'.
        If None, the default path will be used.
    """

    name: str = "ml-driven rss"
    path_to_default_config_parameters: Path | str | None = None

    @job
    def make(self, config_file: str | None = None, **kwargs):
        """
        Make a rss workflow using the specified configuration file and additional keyword arguments.

        Parameters
        ----------
        config_file: str | None
            Path to the configuration file that defines the setup parameters for the whole RSS workflow.
            If not provided, the default file 'rss_default_configuration.yaml' will be used.
        kwargs: dict, optional
            Additional optional keyword arguments to customize the job execution.

        Keyword Arguments
        -----------------
        tag: str
            Tag of systems. It can also be used for setting up elements and stoichiometry.
            For example, the tag of 'SiO2' will be recognized as a 1:2 ratio of Si to O and
            passed into the parameters of buildcell. However, note that this will be overwritten
            if the stoichiometric ratio of elements is defined in the 'buildcell_options'.
        train_from_scratch: bool
            If True, it starts the workflow from scratch.
            If False, it resumes from a previous state.
        resume_from_previous_state: dict | None
            A dictionary containing the state information required to resume a previously interrupted
            or saved RSS workflow. When 'train_from_scratch' is set to False, this parameter is mandatory
            for the workflow to pick up from a saved state.
            Expected keys within this dictionary:
                test_error: float
                    The test error from the last completed training step.
                pre_database_dir: str
                    Path to the directory containing the pre-existing database for resuming.
                mlip_path: str
                    Path to the file of a previous MLIP model.
                isolated_atom_energies: dict
                    A dictionary of isolated atom energy values, with atomic numbers as keys
                    and their energies as valuables.
        generated_struct_numbers: list[int]
            Expected number of generated randomized unit cells by buildcell.
        buildcell_options: list[dict] | None
            Customized parameters for buildcell. Default is None.
        fragment: Atoms | list[Atoms] | None
            Fragment(s) for random structures, e.g., molecules, to be placed indivudally intact.
            atoms.arrays should have a 'fragment_id' key with unique identifiers for each fragment if in same Atoms.
            atoms.cell must be defined (e.g., Atoms.cell = np.eye(3)*20).
        fragment_numbers: list[str] | None
            Numbers of each fragment to be included in the random structures. Defaults to 1 for all specified.
        num_processes_buildcell: int
            Number of processes to use for parallel computation during buildcell generation.
        num_of_initial_selected_structs: list[int] | None
            Number of structures to be sampled directly from the buildcell-generated randomized cells.
        num_of_rss_selected_structs: int
            Number of structures to be selected from each RSS iteration.
        initial_selection_enabled: bool
            If true, sample structures from initially generated randomized cells using CUR.
        rss_selection_method: str
            Method for selecting samples from the RSS trajectories:
            Boltzmann flat histogram in enthalpy first, then CUR.
            Options include:
                bcur1s: Execute bcur with one shot (1s)
                bcur2i: Execute bcur with two iterations (2i)
        bcur_params: dict | None
            Parameters for Boltzmann CUR selection. The default dictionary includes:
                soap_paras: dict
                   SOAP descriptor parameters:
                l_max: int
                    Maximum degree of spherical harmonics (default 12).
                n_max: int
                    Maximum number of radial basis functions (default 12).
                atom_sigma: float
                    Width of Gaussian smearing (default 0.0875).
                cutoff: float
                    Radial cutoff distance (default 10.5).
                cutoff_transition_width: float
                    Width of the transition region (default 1.0).
                zeta: float
                    Exponent for dot-product SOAP kernel (default 4.0).
                average: bool
                    Whether to average the SOAP vectors (default True).
                species: bool
                    Whether to consider species information (default True).
            kb_temp: float
                Temperature in eV for Boltzmann weighting (default 0.3).
            frac_of_bcur: float
                Fraction of Boltzmann CUR selections (default 0.8).
            bolt_max_num: int
                Maximum number of Boltzmann selections (default 3000).
            kernel_exp: float
                Exponent for the kernel (default 4.0).
            energy_label: str
                Label for the energy data (default 'energy').
        random_seed: int | None
            A seed to ensure reproducibility of CUR selection. Default is None.
        include_isolated_atom: bool
            If true, perform single-point calculations for isolated atoms.
        isolatedatom_box: list[float]
            List of the lattice constants for an isolated atom configuration.
        e0_spin: bool
            If true, include spin polarization in isolated atom and dimer calculations. Default is False.
        include_dimer: bool
            If true, perform single-point calculations for dimers only once. Default is False.
        dimer_box: list[float]
            The lattice constants of a dimer box.
        dimer_range: list[float]
            Range of distances for dimer calculations.
        dimer_num: int
            Number of different distances to consider for dimer calculations. Default is 21.
        custom_incar: dict | None
            Dictionary of custom VASP input parameters. If provided, will update the
            default parameters. Default is None.
        custom_potcar: dict | None
            Dictionary of POTCAR settings to update. Keys are element symbols, values are the desired POTCAR labels.
            Default is None.
        vasp_ref_file: str
            Reference file for VASP data. Default is 'vasp_ref.extxyz'.
        config_types: list[str]
            Configuration types for the VASP calculations. Default is None.
        rss_group: list[str]
            Group name for RSS to setting up regularization.
        test_ratio: float
            The proportion of the test set after splitting the data. The value is allowed to be set to 0;
            in this case, the testing error would not be meaningful anymore.
        regularization: bool
            If True, apply regularization. This only works for GAP to date. Default is False.
        scheme: str
            Method to use for regularization. Options are:
                linear_hull: for single-composition system, use 2D convex hull (E, V)
                volume-stoichiometry: for multi-composition system, use 3D convex hull of (E, V, mole-fraction)
        reg_minmax: list[tuple]
            List of tuples of (min, max) values for energy, force, virial sigmas for regularization.
        distillation: bool
            If true, apply data distillation. Default is True.
        force_max: float | None
            Maximum force value to exclude structures. Default is 50.
        force_label: str | None
            The label of force values to use for distillation. Default is 'REF_forces'.
        pre_database_dir: str | None
            Directory where the previous database was saved.
        mlip_type: str
            Choose one specific MLIP type to be fitted: 'GAP' | 'J-ACE' | 'NEQUIP' | 'M3GNET' | 'MACE'.
            Default is 'GAP'.
        ref_energy_name: str
            Reference energy name. Default is 'REF_energy'.
        ref_force_name: str
            Reference force name. Default is 'REF_forces'.
        ref_virial_name: str
            Reference virial name. Default is 'REF_virial'.
        auto_delta: bool
            If true, apply automatic determination of delta for GAP terms. Default is False.
        num_processes_fit: int
            Number of processes used for fitting. Default is 1.
        device_for_fitting: str
            Device to be used for model fitting, either "cpu" or "cuda".
        **fit_kwargs:
            Additional keyword arguments for the MLIP fitting process.
        scalar_pressure_method: str
            Method for adding external pressures.
            Acceptable options are:
                exp: Applies pressure using an exponential distribution.
                uniform: Applies pressure using a uniform distribution.
        scalar_exp_pressure: float
            Scalar exponential pressure. Default is 100.
        scalar_pressure_exponential_width: float
            Width for scalar pressure exponential. Default is 0.2.
        scalar_pressure_low: float
            Low limit for scalar pressure. Default is 0.
        scalar_pressure_high: float
            High limit for scalar pressure. Default is 50.
        max_steps: int
            Maximum number of steps for relaxation. Default is 200.
        force_tol: float
            Force residual tolerance for relaxation. Default is 0.05.
        stress_tol: float
            Stress residual tolerance for relaxation. Default is 0.05.
        hookean_repul: bool
            If true, apply Hookean repulsion. Default is False.
        hookean_paras: dict[tuple[int, int], tuple[float, float]] | None
            Parameters for Hookean repulsion as a dictionary of tuples. Default is None.
        keep_symmetry: bool
            If true, preserve symmetry during relaxation. Default is False.
        write_traj: bool
            If true, write trajectory of RSS. Default is True.
        num_processes_rss: int
            Number of processes used for running RSS. Default is 1.
        device_for_rss: str
            Specify device to use "cuda" or "cpu" for running RSS. Default is "cpu".
        stop_criterion: float
            Convergence criterion for stopping RSS iterations. Default is 0.01.
        max_iteration_number: int
            Maximum number of RSS iterations to perform. Default is 25.
        num_groups: int
            Number of structure groups, used for assigning tasks across multiple nodes.
            For example, if there are 10,000 trajectories to relax and 'num_groups=10',
            the trajectories will be divided into 10 groups and 10 independent jobs will be created,
            with each job handling 1,000 trajectories.
        initial_kb_temp: float
            Initial temperature (in eV) for Boltzmann sampling. Default is 0.3.
        current_iter_index: int
            Index for the current RSS iteration. Default is 1.

        Returns
        -------
        dict
            A dictionary whose keys contains:
                test_error: float
                    The test error of the fitted MLIP.
                pre_database_dir: str
                    The directory of the latest RSS database.
                mlip_path: str
                    The path to the latest fitted MLIP.
                isolated_atom_energies: dict
                    The isolated energy values.
                current_iter: int
                    The current iteration index.
                kb_temp: float
                    The temperature (in eV) for Boltzmann sampling.
        """
        rss_default_config_path = (
            self.path_to_default_config_parameters
            or Path(__file__).absolute().parent / "rss_default_configuration.yaml"
        )

        yaml = YAML(typ="safe", pure=True)

        with open(rss_default_config_path) as f:
            config = yaml.load(f)

        if config_file and os.path.exists(config_file):
            with open(config_file) as f:
                new_config = yaml.load(f)
                config.update(new_config)

        config.update(kwargs)
        self._process_hookean_paras(config)

        config_params = config.copy()

        if "train_from_scratch" not in config_params:
            raise ValueError(
                "'train_from_scratch' must be set in the configuration file or passed as a keyword argument!!"
            )

        rss_flow = []

        if config_params["train_from_scratch"]:
            initial_exclude_keys = [
                "train_from_scratch",
                "resume_from_previous_state",
                "config_types",
                "rss_group",
                "num_of_rss_selected_structs",
                "rss_selection_method",
                "scalar_pressure_method",
                "scalar_exp_pressure",
                "scalar_pressure_exponential_width",
                "scalar_pressure_low",
                "scalar_pressure_high",
                "max_steps",
                "force_tol",
                "stress_tol",
                "stop_criterion",
                "max_iteration_number",
                "num_groups",
                "initial_kb_temp",
                "current_iter_index",
                "hookean_repul",
                "hookean_paras",
                "keep_symmetry",
                "write_traj",
                "num_processes_rss",
                "device_for_rss",
            ]
            initial_params = {
                k: v for k, v in config_params.items() if k not in initial_exclude_keys
            }
            initial_params.update(
                {
                    "config_type": config_params["config_types"][0],
                    "rss_group": config_params["rss_group"][0],
                }
            )
            initial_rss_job = initial_rss(**initial_params)
            rss_flow.append(initial_rss_job)

        rss_group = config_params["rss_group"]
        config_types = config_params["config_types"]
        do_rss_group = rss_group[0] if len(rss_group) == 1 else rss_group[-1]
        rss_config_type = (
            config_types[0] if len(config_types) == 1 else config_types[1:]
        )

        rss_exclude_keys = [
            "train_from_scratch",
            "resume_from_previous_state",
            "pre_database_dir",
        ]

        rss_params = {
            k: v for k, v in config_params.items() if k not in rss_exclude_keys
        }
        rss_params.update(
            {
                "num_of_initial_selected_structs": None,
                "initial_selection_enabled": False,
                "rss_group": do_rss_group,
                "config_types": rss_config_type,
            }
        )

        if config_params["train_from_scratch"]:
            rss_params.update({"include_isolated_atom": False})
            rss_params.update({"include_dimer": False})

            do_rss_job = do_rss_iterations(
                input=initial_rss_job.output,
                **rss_params,
            )
        else:
            if "resume_from_previous_state" not in config_params:
                raise ValueError(
                    "The parameter 'resume_from_previous_state' must be specified when 'train_from_scratch' is False."
                )
            resume_from_previous_state = config_params["resume_from_previous_state"]
            do_rss_job = do_rss_iterations(
                input=resume_from_previous_state,
                **rss_params,
            )

        rss_flow.append(do_rss_job)

        return Response(replace=Flow(rss_flow), output=do_rss_job.output)

    def _process_hookean_paras(self, config):
        if "hookean_paras" in config:
            config["hookean_paras"] = {
                tuple(map(int, k.strip("()").split(", "))): tuple(v)
                for k, v in config["hookean_paras"].items()
            }
